
#include <stdio.h>
#include <windows.h>
#include <snanls.h>
#include <process.h>
#include <string.h>

#define OTHEXPORT __declspec(dllexport)

#include "oh_trmod.h"

#define BUFFSIZE 2048
#define TRANSBUFFSIZE ((BUFFSIZE) * 4)

#define OTHLOCKINIT() InitializeCriticalSection(&CriticalSection)
#define OTHLOCK() EnterCriticalSection(&CriticalSection)
#define OTHUNLOCK() LeaveCriticalSection(&CriticalSection)
#define OTHLOCKDEL() DeleteCriticalSection(&CriticalSection)

static unsigned int iJKTransServerAddress;
static unsigned int iJKTransServerPort = 80;
static char *lpszJKTransServerName = "kr.japan.yahoo.com";
/* static char *lpszJKTransServerName = "japan.empas.com"; */
static char *lpszJKTransQueryPrefix =
  "GET /translation/trans_stext.html?trans=j2k&text=";

static HINSTANCE hInst;
static int isRunning;

static CRITICAL_SECTION CriticalSection;
static char utfjpbuff[BUFFSIZE];
static int utfjpendpos;
static char euckrbuff[BUFFSIZE];
static int euckrendpos;
static HANDLE jktrth;

static char *OthJKTrans(char *utfjp);
static void OthPushUTFJP(char *utfjp, int len);
static char *OthPullUTFJP(void);
static void OthPushEUCKR(char *euckr);
static char *OthPullEUCKR(void);

static void OthJKTransThread(void *arg);

static char *OthJKTrans(char *utfjp)
{
  struct sockaddr_in sock;
  int sd;
  char transbuf[TRANSBUFFSIZE];
  int transbufsize = sizeof(transbuf);
  char *recvbuf;
  int recvbufsize;
  int i, j;
  int timeout = 10000;

  memset(&sock, 0, sizeof(sock));
  sock.sin_family = AF_INET;
  sock.sin_addr.s_addr = iJKTransServerAddress;
  sock.sin_port = htons((unsigned short)iJKTransServerPort);

  if ((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
  {
    return NULL;
  }

  if (setsockopt(sd, SOL_SOCKET, SO_RCVTIMEO, (const char *)&timeout,
    sizeof(timeout)) == SOCKET_ERROR)
  {
    closesocket(sd);
    return NULL;
  }

  if (connect(sd, (struct sockaddr *)&sock, sizeof(sock)) == -1)
  {
    closesocket(sd);
    return NULL;
  }

  strcpy(transbuf, lpszJKTransQueryPrefix);
  j = strlen(transbuf);

  for (i = 0; utfjp[i] != '\0' && transbufsize > j + 16; ++i)
  {
    sprintf(&transbuf[j], "%%%02x", (unsigned char)utfjp[i]);
    j += 3;
  }
  transbuf[j] = '\x0a';
  ++j;
  transbuf[j] = '\0';

  if (send(sd, transbuf, strlen(transbuf), 0) == -1)
  {
    closesocket(sd);
    return NULL;
  }

  recvbuf = malloc(TRANSBUFFSIZE);
  recvbufsize = TRANSBUFFSIZE;

  i = 0;

  while ((j = recv(sd, &recvbuf[i], TRANSBUFFSIZE, 0)) > 0)
  {
    i += j;
    if (i + TRANSBUFFSIZE > recvbufsize)
    {
      recvbuf = realloc(recvbuf, (recvbufsize = recvbufsize + TRANSBUFFSIZE));
      if (recvbuf == NULL)
      {
        closesocket(sd);
        return NULL;
      }
    }
  }

  closesocket(sd);

  j = i = strlen(recvbuf);
  --i;

  for (; recvbuf[i] != '>' && i >= 0; --i)
  {
  }

  if (i == -1)
  {
    free(recvbuf);
    return NULL;
  }

  memmove(recvbuf, &recvbuf[i + 1], j - i);

  recvbuf = realloc(recvbuf, j - i);

  return recvbuf;
}

static void OthPushUTFJP(char *utfjp, int len)
{
  int i;

  if (len <= 0)
    return;

  OTHLOCK();

  if (utfjpendpos >= BUFFSIZE)
  {
    utfjpbuff[0] = '\0';
    utfjpendpos = 0;
  }

  for (i = 0; i < len && utfjpendpos < BUFFSIZE && utfjp[i] != '\0';
    ++i, ++utfjpendpos)
  {
    utfjpbuff[utfjpendpos] = utfjp[i];
  }

  if (utfjpendpos >= BUFFSIZE)
  {
    utfjpbuff[0] = '\0';
    utfjpendpos = 0;
  }

  utfjpbuff[utfjpendpos] = '\0';

  OTHUNLOCK();
}

static const char *delimiters[] =
  {
    "\xef\xbc\x8e",
    "\x2e",
    "\xe3\x80\x81",
    "\xef\xbd\xa4",
    "\xef\xbc\x8c",
    "\x2c",
    "\xef\xbc\x9f",
    "\x3f",
    "\xe2\x80\xa6",
    "\xe3\x83\xbb",
    "\xef\xbc\x81",
    "\xef\xbd\xa5",
    "\x21",
  };
static const int ndelimiters = sizeof(delimiters) / sizeof(char *);

static char *OthPullUTFJP(void)
{
  int dpos, newdpos, i;
  char *newdposptr;
  char *retutfjp;

  OTHLOCK();

  if (utfjpendpos == 0)
  {
    OTHUNLOCK();
    return NULL;
  }

  for (i = 0, dpos = utfjpendpos; i < ndelimiters; i++)
  {
    newdposptr = strstr(utfjpbuff, delimiters[i]);
    if (newdposptr != NULL)
    {
      newdpos = newdposptr - utfjpbuff + strlen(delimiters[i]);
      if (dpos > newdpos)
      {
        dpos = newdpos;
      }
    }
  }

  if (dpos == utfjpendpos)
  {
    retutfjp = strdup(utfjpbuff);
    utfjpbuff[0] = '\0';
    utfjpendpos = 0;
  }
  else
  {
    retutfjp = malloc(newdpos + 1);
    memcpy(retutfjp, utfjpbuff, newdpos);
    retutfjp[newdpos] = '\0';
    memmove(utfjpbuff, utfjpbuff + newdpos, strlen(utfjpbuff + newdpos) + 1);
    utfjpendpos = strlen(utfjpbuff);
  }

  for (i = 0; i < ndelimiters; i++)
  {
    if (strcmp(delimiters[i], retutfjp) == 0)
    {
      retutfjp[0] = '\0';
      break;
    }
  }

  OTHUNLOCK();

  return retutfjp;
}

static void OthPushEUCKR(char *euckr)
{
  int i, len;

  if (euckr == NULL)
    return;

  len = strlen(euckr);

  OTHLOCK();

  if (euckrendpos >= BUFFSIZE)
  {
    euckrbuff[0] = '\0';
    euckrendpos = 0;
  }

  for (i = 0; i < len && euckrendpos < BUFFSIZE; ++i, ++euckrendpos)
  {
    euckrbuff[euckrendpos] = euckr[i];
  }

  if (euckrendpos >= BUFFSIZE)
  {
    euckrbuff[0] = '\0';
    euckrendpos = 0;
  }

  euckrbuff[euckrendpos] = '\0';

  OTHUNLOCK();
}

static char *OthPullEUCKR(void)
{
  char *ret = NULL;

  OTHLOCK();

  if (euckrendpos != 0)
  {
    ret = strdup(euckrbuff);
    euckrbuff[0] = '\0';
    euckrendpos = 0;
  }

  OTHUNLOCK();

  return ret;
}

static void OthJKTransThread(void *arg)
{
  char *utfjp;
  char *euckr;

  while (isRunning)
  {
    utfjp = OthPullUTFJP();
    if (utfjp != NULL)
    {
      if (utfjp[0] != '\0')
      {
        euckr = OthJKTrans(utfjp);
        OthPushEUCKR(euckr);
        free(euckr);
      }
      free(utfjp);
    }
    else
    {
      Sleep(3000);
    }
  }
}

OTHEXPORT int OTHCALL OthTransInit(OTHTR *pTR)
{
  struct hostent *hp;
  WSADATA wsaData;

  if (isRunning)
  {
    ++isRunning;
    return 0;
  }

  pTR->lpStr = NULL;
  pTR->iLen = 0;

  WSAStartup(MAKEWORD(2, 0), &wsaData);

  hp = gethostbyname(lpszJKTransServerName);

  if (hp == 0)
    return -1;

  iJKTransServerAddress = ((struct in_addr *)(hp->h_addr))->s_addr;

  OTHLOCKINIT();

  isRunning = 1;

  jktrth = (HANDLE)_beginthread(&OthJKTransThread, 0, NULL);

  return 0;
}

OTHEXPORT int OTHCALL OthTransSetting(OTHTR *pTR)
{
  return 0;
}

OTHEXPORT int OTHCALL OthTransCopySetting(OTHTR *pDestTR, OTHTR *pSrcTR)
{
  return 0;
}

OTHEXPORT int OTHCALL OthTranslate(OTHTR *pTR, LPCWSTR lpStr, int iLen)
{
  int len;
  LPSTR euckr;

  if (lpStr[0] != 0)
  {
    LPSTR utfjp;
    len = WideCharToMultiByte(CP_UTF8, 0, lpStr, iLen, 0, 0, 0, 0);
    utfjp = _alloca(len);
    WideCharToMultiByte(CP_UTF8, 0, lpStr, iLen, utfjp, len, 0, 0);
    OthPushUTFJP(utfjp, len);
  }

  euckr = OthPullEUCKR();

  free(pTR->lpStr);
  pTR->lpStr = NULL;

  if (euckr)
  {
    len = MultiByteToWideChar(CP_949, 0, euckr, -1, 0, 0);
    pTR->lpStr = malloc(len << 1);
    pTR->iLen = len;
    MultiByteToWideChar(CP_949, 0, euckr, -1, pTR->lpStr, len);
  }

  if (pTR->lpStr == NULL)
  {
    pTR->lpStr = malloc(2);
    pTR->lpStr[0] = 0;
    pTR->iLen = 0;
  }

  return 0;
}

OTHEXPORT int OTHCALL OthTransDestroy(OTHTR *pTR)
{
  --isRunning;
  if (isRunning != 0)
    return 0;

  if (jktrth)
  {
    WaitForSingleObject(jktrth, INFINITE);
    jktrth = NULL;
  }
  utfjpbuff[0] = '\0';
  utfjpendpos = 0;
  euckrbuff[0] = '\0';
  euckrendpos = 0;
  OTHLOCKDEL();
  free(pTR->lpStr);

  return 0;
}

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
    hInst = hinstDLL;

  return TRUE;
}
